DEFAULT_SPIDER_GLOBAL_CONFIG = {
    'spider_modules': ['spider'],
    'save_report': True,
    'thread_number': 1,
    'network_try_limit': 3,
    'task_try_limit': 3,
    'display_timing': False,
    'display_stats': True,
}
